import numpy as np
import pandas as pd
import stan
import pickle
import arviz as az


StanCode = """
    //
    //          Yasuharu Okamoto, 2023.08, 2025.03
    //
    data {
        int N;
        int M;
        array[N,M] real X;
        vector[M] i_mu;
        vector[M] i_std;
    }
    parameters {
        vector[N] vF;
        vector[M] v_mu;
        vector<lower=0.0>[M] v_Lmbd;
        vector<lower=0.0>[M] v_sgm;
    }
    transformed parameters {
        vector[N] F;
        real sumF;
        real ssF;
        real meanF;
        real sdF;
        vector[M] mu;
        vector[M] Lmbd;
        vector[M] sgm;
        array[N] vector[M] V;
        //
        //        Standardized F values
        //
        sumF = 0.0;
        ssF = 0.0;
        for (i in 1:N) {
            sumF += vF[i];
            ssF += pow(vF[i], 2);
        }
        meanF = sumF / N;
        sdF = pow((ssF - N * pow(meanF, 2)) / (N - 1), 0.5);
        F = (vF - meanF) / sdF;       //  標準化
        //
        //          Rescaled parameters
        //
        mu = i_mu + 10.0 .* i_std .* v_mu;
        Lmbd = 10.0 .* i_std .* v_Lmbd;
        sgm = 0.001 + 10.0 .* i_std .* v_sgm;
        //
        //          Regression model
        //
        for (i in 1:N)
            V[i] = mu + Lmbd * F[i];
    }
    model {
        vF ~ normal(0.0, 1.0);
        v_mu ~ normal(0.0, 10.0);
        v_Lmbd ~ exponential(0.1);
        v_sgm ~ exponential(0.1);
        for (i in 1:N)
            X[i] ~ normal(V[i], sgm);
    }
    generated quantities {
        real omg;               //  omega
        vector[M] L2V;
        real v1;
        real v2;
        v1 = 0.0;
        v2 = 0.0;
        L2V = pow(Lmbd, 2.0) ./ (pow(Lmbd, 2.0) + pow(sgm, 2.0));
        for (j in 1:M) {
            v1 += Lmbd[j];
            v2 += pow(sgm[j], 2);
        }
        omg = pow(v1, 2) / (pow(v1, 2) + v2);
    }

"""

pd.set_option('display.max_columns', 10000)
pd.set_option('display.max_rows', 10000)

fout = open('Results.txt', 'w')
#
#       Excelデータファイルの読み込み
#
s = input("Input Excel data file (*.xlsx) = ")
xlsx_fl = pd.ExcelFile(s)
data_xlsx = pd.read_excel(xlsx_fl)

with open('data.pkl', 'wb') as f:
    pickle.dump(data_xlsx, f)

var_names = [v for v in data_xlsx.keys()][1:]
print('var_names =', var_names)

data = data_xlsx.values
print('data...\n', data)
fout.write(f'Input data file = {s}\n')
fout.write(f'\ndata...\n{data}\n')
X = data[:,1:]

mus = np.mean(X, axis=0)
stds = np.std(X, axis=0)

Data = {'N': len(X), 'M': len(X[0]), 'X': X, 'i_mu': mus, 'i_std': stds}

sm = stan.build(StanCode, data = Data)

fit = sm.sample()

df = fit.to_frame()            #   DataFrame型への変換

with open('fit.pkl', 'wb') as f:
    pickle.dump(df, f)

print()
fout.write('\n\nMu...\n')
for j in range(len(X[0])):
    q1, med, q3 = np.percentile(df[f'mu.{j+1}'], [25, 50, 75])
    print(f'Mu[{j+1}]:', end = '')
    print(f'  Q1 = {q1:.3f},  Med = {med:.3f},  Q3 = {q3:.3f}')
    fout.write(f'Mu[{j+1}]:  Q1 = {q1:.3f},  Med = {med:.3f},  Q3 = {q3:.3f}\n')

i_data = az.from_pystan(posterior=fit, posterior_model=sm)
with open('i_data.pkl', 'wb') as f:
    pickle.dump(i_data, f)         #   InferenceData of MCMC samples is saved
print(az.summary(i_data))          #   Ptint summary of MCMC sample
fout.write(f'\n{az.summary(i_data)}\n')
              
fout.close()
print('\nResults.txt was saved.\n')
